{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "8bdc762b",
   "metadata": {},
   "source": [
    "# Week 7 - Monte-Carlo, Block Averages, NumExpr and Metropolis Algorithm"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "49409b74",
   "metadata": {},
   "source": [
    "## From Monte-Carlo to Block Averages\n",
    "\n",
    ">Let's write a code that calculates $\\pi$ with Monte-Carlo sampling and block averages.\n",
    "\n",
    "We will employ a uniform random distribution $x \\in [0,1]$ to calculate $\\pi$ with the following integral:\n",
    "\n",
    "$$ 4\\int_0^1 \\sqrt{1-x^2} dx$$"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "ab18469b",
   "metadata": {},
   "outputs": [],
   "source": [
    "import numpy as np\n",
    "import matplotlib.pyplot as plt\n",
    "import numexpr as ne\n",
    "\n",
    "class Integral:\n",
    "    '''\n",
    "    This class calculates \\pi \n",
    "    through Monte Carlo integration\n",
    "    '''\n",
    "    \n",
    "    def __init__(self,N,numexp=False):\n",
    "        'Initialise the MC.'\n",
    "        self.N = N\n",
    "        self.rng = np.random.default_rng(seed=42424)\n",
    "        self.I = 0.0\n",
    "        self.I2 = 0.0\n",
    "        self.stdvn = 0.0\n",
    "        self.do_block_averages = None\n",
    "        self.blocks = None\n",
    "        self.var = None\n",
    "        self.numexp = numexp\n",
    "        \n",
    "    def f(self,x):\n",
    "        'Function to integrate.'\n",
    "        return 4.0*np.sqrt(1.0-x**2)\n",
    "    \n",
    "    def f2(self,x):\n",
    "        'Square of self.f, computationally faster.'\n",
    "        return 16.0*(1.0-x**2)\n",
    "    \n",
    "    def compute_mc_integral(self,x):\n",
    "        'Sample self.f and average.'\n",
    "        if self.numexp:\n",
    "            ev = ne.evaluate(\"(4.0*sqrt(1.0-x**2))\",optimization='moderate')\n",
    "            return np.average(ev)\n",
    "        else:\n",
    "            return sum(map(self.f, x))/len(x)\n",
    "\n",
    "    def compute_mc_integral_var(self,x):\n",
    "        '''\n",
    "        Sample self.f and calculates\n",
    "        mean, mean of the square, variance,\n",
    "        standard deviation\n",
    "        '''\n",
    "        I = self.compute_mc_integral(x)\n",
    "        I2 = sum(map(self.f2, x))/len(x)\n",
    "        var = (I2 - I**2)\n",
    "        stdvn =  np.sqrt((I2 - I**2)/len(x))\n",
    "        return np.array([I,I2,stdvn,var])\n",
    "        \n",
    "    def mc_integral(self,block_averages=False,blocks=1):\n",
    "        'MC integration driver'\n",
    "        self.do_block_averages = block_averages\n",
    "        self.blocks = blocks\n",
    "        r = self.rng.random(self.N)\n",
    "        if self.do_block_averages and self.blocks!=self.N:\n",
    "            rs = np.split(r,self.blocks)\n",
    "            results = np.array(list(map(self.compute_mc_integral,rs)))\n",
    "            self.I = np.average(results)\n",
    "            self.I2 =  np.average(results**2)\n",
    "            self.stdvn = np.sqrt((self.I2 - self.I**2)/results.shape[0])\n",
    "            self.var = (self.I2 - self.I**2)\n",
    "        else:\n",
    "            self.I,self.I2,self.stdvn,self.var = self.compute_mc_integral_var(r)\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "de136598",
   "metadata": {},
   "outputs": [],
   "source": [
    "integral = Integral(10000)\n",
    "integral.mc_integral()\n",
    "for i in range(5):\n",
    "    integral.mc_integral()\n",
    "    print(integral.I)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "6610c3c3",
   "metadata": {},
   "outputs": [],
   "source": [
    "y = []\n",
    "yy = []\n",
    "x = range(100,200000,1000)\n",
    "for N in x:\n",
    "    integral = Integral(N)\n",
    "    integral.mc_integral()\n",
    "    y.append(integral.I)\n",
    "    yy.append(abs(integral.I - np.pi))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "d9604d09",
   "metadata": {},
   "outputs": [],
   "source": [
    "plt.plot(x,y,'o-')\n",
    "plt.axhline(np.pi,c='k',alpha=0.5)\n",
    "plt.ylabel('integral')\n",
    "plt.xlabel('N')"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "84e77299",
   "metadata": {},
   "source": [
    "Now we check whether the true error goes down as $1/\\sqrt{N}$."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "a3f61812",
   "metadata": {},
   "outputs": [],
   "source": [
    "plt.loglog(x,yy,'o-',label='MC integral error')\n",
    "plt.loglog(x,list(map(lambda x: 1/np.sqrt(x),x)),label='$\\\\frac{1}{\\sqrt{N}}$')\n",
    "plt.ylabel('absolute error')\n",
    "plt.xlabel('N')\n",
    "plt.legend()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "dac8f342",
   "metadata": {},
   "outputs": [],
   "source": [
    "# Whole data set\n",
    "N=10000\n",
    "integral = Integral(N)\n",
    "integral.mc_integral()\n",
    "print('I = {:.5f}, variance = {:.5f}, stdvn = {:.5f}, Delta= {:.5f}'.\n",
    "      format(integral.I,integral.var, integral.stdvn, abs(integral.I-np.pi)))\n",
    "print('-'*20)\n",
    "print('sigma_n = {}'.format(integral.stdvn*np.sqrt(N)))\n",
    "print('sigma_n/sqrt(N) = {}'.format(integral.stdvn))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "d06fad6c",
   "metadata": {},
   "outputs": [],
   "source": [
    "#Bloch averages: method 1\n",
    "N = int(1e4)\n",
    "l_I = []\n",
    "l_I2 = []\n",
    "integral = Integral(N)\n",
    "NN = 10\n",
    "for i in range(NN):\n",
    "    integral.mc_integral()\n",
    "    print('I = {:.5f}, variance = {:.5f}, stdvn = {:.5f}, Delta= {:.5f}'.format(integral.I,integral.var, \n",
    "                                                                integral.stdvn, abs(integral.I-np.pi)))\n",
    "    l_I.append(integral.I)\n",
    "    l_I2.append(integral.I**2)\n",
    "var = np.average(l_I2)-(np.average(l_I))**2\n",
    "sigma_m = np.sqrt(var)\n",
    "print('-'*20)\n",
    "print('sigma_m = {:.5f}'.format(sigma_m))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "3048138c",
   "metadata": {},
   "outputs": [],
   "source": [
    "#Bloch averages: method 2\n",
    "N=10000\n",
    "integral = Integral(N)\n",
    "integral.mc_integral(block_averages=True,blocks=10)\n",
    "print('I = {:.5f}, variance = {:.5f}, stdvn = {:.5f}, Delta= {:.5f}'.\\\n",
    "      format(integral.I,integral.var, integral.stdvn, abs(integral.I-np.pi)))\n",
    "print('-'*20)\n",
    "print('sigma_s/sqrt(s) = {:.5f}'.format(integral.stdvn))"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "52db3a9c",
   "metadata": {},
   "source": [
    "-------------"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "47704a7b",
   "metadata": {},
   "source": [
    "# NumExpr\n",
    "\n",
    "From the project description on PyPI https://pypi.org/project/numexpr/:\n",
    "\n",
    ">**NumExpr is a fast numerical expression evaluator for NumPy**. \n",
    ">\n",
    ">With it, expressions that operate on arrays (like '3*a+4*b') are accelerated and use less memory than doing the same calculation in Python.\n",
    ">\n",
    ">In addition, its multi-threaded capabilities can make use of all your cores – which generally results in substantial performance scaling compared to NumPy.\n",
    ">\n",
    ">Last but not least, numexpr can make use of Intel’s VML (Vector Math Library, normally integrated in its Math Kernel Library, or MKL). This allows further acceleration of transcendent expressions"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "eb430130",
   "metadata": {},
   "outputs": [],
   "source": [
    "%load_ext line_profiler\n",
    "integral = Integral(int(1e6))\n",
    "%lprun -f Integral.mc_integral integral.mc_integral(block_averages=True,blocks=100)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "655bc688",
   "metadata": {},
   "outputs": [],
   "source": [
    "%lprun -f Integral.compute_mc_integral integral.mc_integral(block_averages=True,blocks=100)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "719eb358",
   "metadata": {},
   "outputs": [],
   "source": [
    "integral = Integral(int(1e6),numexp=False)\n",
    "%timeit integral.mc_integral(block_averages=True,blocks=10)\n",
    "print(integral.I,integral.stdvn)\n",
    "integral = Integral(int(1e6),numexp=True)\n",
    "%timeit integral.mc_integral(block_averages=True,blocks=10)\n",
    "print(integral.I,integral.stdvn)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "f5e0b57a",
   "metadata": {},
   "source": [
    "Note the speed-up of two orders of magnitued obtained by using ``numexpr``!"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "c1774d30",
   "metadata": {},
   "outputs": [],
   "source": [
    "integral = Integral(int(1e6),numexp=True)\n",
    "%lprun -f Integral.compute_mc_integral integral.mc_integral(block_averages=True,blocks=100)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "4bb209ad",
   "metadata": {},
   "source": [
    "> The ``NumExpr`` package provides tools for fast evaluation of array expressions. \n",
    "It is particularly useful for **large arrays**.\n",
    "\n",
    "The key to NumExpr’s speed is handling chunks of elements at a time. Default is 4096 elements at a time, using a register machine. NumExpr \"sort of\" compiles the expression in a vectorized manner and also support multi-threading parallelization by default. Check out the docs at: https://numexpr.readthedocs.io/en/latest/index.html\n"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "a77d2528",
   "metadata": {},
   "source": [
    "-----------------------------"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "905d44c9",
   "metadata": {},
   "source": [
    "## Metropolis Algorithm\n",
    "\n",
    "Now we will focus on the **Metropolis algorithm**, which has been introduced to you in the lecture of yesterday. In particular, we will sample the distribution\n",
    "\n",
    "$$P(x) = e^{\\frac{-x^2}{2\\sigma^2}}.$$\n",
    "\n",
    "In order to do so, we will create a Python class that runs a Metropolis algorithm reproducing the distribution $P(x)$. We will employ a **Markov chain** obtained through **one random walker** behaving according to the following transition probability:\n",
    "\n",
    "$$T(x_i \\rightarrow x_j) = \\min{\\left[1,P(x_j)/P(x_i)\\right]},$$\n",
    "\n",
    "where the trial move is obtained with a uniformly distributed random variable $\\tilde{x} \\in [0,1) $\n",
    "\n",
    "$$ x_j = x_i + \\delta(\\tilde{x}-0.5) $$\n",
    "\n",
    "which is accepted if \n",
    "\n",
    "$$\\tilde{x}_w \\leq w = P(x_j)/P(x_i) = e^{\\frac{-(x_j^2-x_i^2)}{2\\sigma^2}} $$\n",
    "\n",
    "with $\\tilde{x}_w \\in [0,1)$ also being a random variable following a uniform distribution."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "9ee1bf5e",
   "metadata": {},
   "outputs": [],
   "source": [
    "import numpy as np\n",
    "import matplotlib.pyplot as plt\n",
    "\n",
    "class Metropolis:\n",
    "    '''\n",
    "    This class implements \n",
    "    the Metropolis algorithm for a Gaussian distribution\n",
    "    exp(-x**2/(2*sigma**2))\n",
    "    '''\n",
    "    \n",
    "    def __init__(self,N,delta,sigma=1):\n",
    "        'Initialise the Metropolis'\n",
    "        self.N = N\n",
    "        self.rng = np.random.default_rng(seed=42424)\n",
    "        self.sigma = sigma\n",
    "        self.delta = delta\n",
    "        self.pref = -1.0/(2.0*self.sigma**2)\n",
    "        self.acc = 0\n",
    "        self.x = 0.0\n",
    "        self.ar = None\n",
    "        self.all_x = [self.x]\n",
    "        self.all_x2 = None\n",
    "        self.all_x3 = None\n",
    "        self.all_x4 = None\n",
    "        self.var = None\n",
    "        self.counter = 0\n",
    "        self.autocorr = []\n",
    "        self.n_autocorr = 10\n",
    "    \n",
    "    def metro_step(self):\n",
    "        'One step of Metropolis'\n",
    "        rnd = self.rng.random()\n",
    "        xp = self.x + self.delta * (rnd-0.5)\n",
    "        p = self.x**2\n",
    "        pp = xp**2\n",
    "        w = np.exp(self.pref*(pp-p))\n",
    "        rnd_w = self.rng.random()\n",
    "        if (rnd_w<=w):\n",
    "            self.x = xp \n",
    "            self.acc += 1\n",
    "        self.all_x.append(self.x)\n",
    "    \n",
    "    def metro_driver(self):\n",
    "        'Run N Metropolis steps.'\n",
    "        for i in range(self.N):\n",
    "            self.counter += 1\n",
    "            self.metro_step()\n",
    "    \n",
    "    def get_hist(self):\n",
    "        'Plot Histogram'\n",
    "        plt.xlabel('x')\n",
    "        plt.ylabel('count')\n",
    "        plt.hist(self.all_x,bins=100)\n",
    "        plt.show()\n",
    "    \n",
    "    @staticmethod\n",
    "    def calc_autocorr(a,N):\n",
    "        'Autocorrelation function'\n",
    "        res = np.ones(N)\n",
    "        for k in range(1,N):\n",
    "            c_k = np.average(a[:-k]*a[k:]) \n",
    "            x_r = np.average(a[:-k])\n",
    "            x_r_2 = np.average(a[:-k]**2)\n",
    "            res[k] = (c_k-x_r**2)/(x_r_2-x_r**2)\n",
    "        return res\n",
    "    \n",
    "    def get_stats(self):\n",
    "        '''\n",
    "        Calculate, store and print \n",
    "        statistical quantities\n",
    "        '''\n",
    "        self.ar = self.acc*1.0/self.counter\n",
    "        self.aver_x = np.average(self.all_x)\n",
    "        self.aver_x2 = np.average(np.array(self.all_x)**2)\n",
    "        self.aver_x3 = np.average(np.array(self.all_x)**3)\n",
    "        self.aver_x4 = np.average(np.array(self.all_x)**4)\n",
    "        self.var = self.aver_x2-self.aver_x**2\n",
    "        print(\"Total number of steps run = {}\".format(self.counter))\n",
    "        print(\"Acceptance ratio = {}\".format(self.ar))\n",
    "        print(\"Sigma = {}, delta = {}\".format(self.sigma,self.delta))\n",
    "        print(\"<x> = {}, <x2> = {}\".format(self.aver_x,self.aver_x2))\n",
    "        print(\"<x3> = {}, <x4> = {}\".format(self.aver_x3,self.aver_x4))\n",
    "        print(\"variance = {}\".format(self.var))\n",
    "        \n",
    "    def get_autocorr(self,n_autocorr=100):\n",
    "        '''\n",
    "        Wrapper for the autocorrelation\n",
    "        function calculator\n",
    "        '''\n",
    "        a = np.array(self.all_x)\n",
    "        self.autocorr = self.calc_autocorr(a,n_autocorr)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "5f54de78",
   "metadata": {},
   "source": [
    "Let's run our first Metropolis!"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "78396651",
   "metadata": {},
   "outputs": [],
   "source": [
    "N= 100000\n",
    "delta = 5\n",
    "metro = Metropolis(N,delta)   #Comment to run multiple times\n",
    "metro.metro_driver()\n",
    "metro.get_stats()\n",
    "metro.get_hist()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "51adb017",
   "metadata": {},
   "source": [
    "Let's study the effect of changing N. \\\n",
    "Also, how does the acceptance ratio change if we change $\\sigma$?"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "aa6f0b12",
   "metadata": {},
   "outputs": [],
   "source": [
    "metro = Metropolis(10000,5)   \n",
    "metro.metro_driver()\n",
    "metro.get_stats()\n",
    "metro.get_hist()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "966d07dc",
   "metadata": {},
   "outputs": [],
   "source": [
    "x = np.arange(100,50000,100)\n",
    "y = []\n",
    "for N in x:\n",
    "    metro = Metropolis(N,5)\n",
    "    metro.metro_driver()\n",
    "    metro.get_stats()\n",
    "    y.append(1.0-metro.var)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "3572f8aa",
   "metadata": {},
   "outputs": [],
   "source": [
    "plt.xlabel('N')\n",
    "plt.ylabel('$\\sigma^2_{true}-\\sigma^2_{calc}$')\n",
    "plt.axhline(0.0,c='grey',alpha=0.5)\n",
    "plt.axhline(0.05,ls='--',c='grey',alpha=0.5)\n",
    "plt.axhline(-0.05,ls='--',c='grey',alpha=0.5)\n",
    "plt.plot(x,y,'o-')\n",
    "plt.show()"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "fis_comp_2023",
   "language": "python",
   "name": "fis_comp_2023"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.11.1"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
